#include <stdio.h>
#include <vector>
#include <ext/hash_map>
#include <assert.h>

#include "reading.h"
#include "data.h"

using namespace std;
using namespace __gnu_cxx;

#define SIZE 2

int serial = 0;

Data *entry_labels[SIZE];

// int live_subs = 0;

struct Substitution : public Data {
    Data *entry[SIZE];
    bool antimatter;
    
    Substitution() {
        for (int i = 0; i < SIZE; i++) {
            entry[i] = NULL;
        }

        antimatter = false;

//        live_subs++;
//        printf("# live_subs %d\n", live_subs);
    }

    ~Substitution() {
//        live_subs--;
    }
    
    void print() {
        printf("<");

        if (antimatter) {
            printf("antimatter ");
        }
        
        for (int i = 0; i < SIZE; i++) {
            if (! entry_labels[i]) break;
            
            if (i) {
                printf(", ");
            }

            entry_labels[i]->print();
            printf(" : ");
            
            if (entry[i]) {
                entry[i]->print();                 
            } else {
                printf("None");
            }
        }

        printf(">");
    }
};

struct hash_substitution {
    size_t operator() (const Substitution *s) const {
        int rv = s->antimatter ? 1 : 0;

        for (int i = 0; i < SIZE; i++) {
            rv = rv * 12347 + (int) s->entry[i];
        }

        return (size_t) rv;
    }
};

struct equal_substitution {
    bool operator() (const Substitution *a, const Substitution *b) const {
        if (a->antimatter != b->antimatter)
            return 0;

        for (int i = 0; i < SIZE; i++) {
            if (a->entry[i] != b->entry[i]) {
                return 0;
            }
        }
 
        return 1;
    }
};

hash_map<Substitution *, Substitution *,
         hash_substitution, equal_substitution> sub_discrim;

static IntReading *One = NULL;
static IntReading *Zero = NULL;
static StringReading *Or = NULL;
static StringReading *Not = NULL;
static StringReading *Wildcard = NULL;
static StringReading *Unreachable = NULL;
static Data *Empty;
static Data *ProgramStart;


static Substitution *empty = NULL;

void init_prpq_code(int argc, char **argv) {
    One = IntReading::factory(1);
    Zero = IntReading::factory(0);
    Or = StringReading::factory("or");
    Not = StringReading::factory("not");
    Wildcard = StringReading::factory("wildcard");
    Unreachable = StringReading::factory("unreachable!");

    Empty = new FunctorReading(StringReading::factory("EMPTY"),
                               new TupleReading());

    ProgramStart = new FunctorReading(StringReading::factory("program_start"),
                                      new TupleReading());
    
    empty = new Substitution();
    sub_discrim[empty] = empty;
}

Data *empty_subst() {
    return empty;
}

Data *empty_label() {
    return Empty;
}

Data *program_start() {
    return ProgramStart;
}

Data *equal(Data *a, Data *b) {
    return a == b ? One : Zero;
}

Data *is_antimatter(Data *sub_d) {
    Substitution *sub = dynamic_cast<Substitution *>(sub_d);

    if (sub->antimatter) {
        return One;
    } else {
        return Zero;
    }
}

Data *and_op(Data *a, Data *b) {
    if (a == Zero) return Zero;
    return b;
}

Data *subst(Data *tl_d, Data *el_d) {

//     printf("TL and EL:");
//     tl_d->print();
//     printf(" ");
//     el_d->print();
//     printf("\n");
    
    static int serial = 0;
   
    if (tl_d == Wildcard) {
        return empty;
    }

    FunctorReading *tl = dynamic_cast<FunctorReading *>(tl_d);
    FunctorReading *el = dynamic_cast<FunctorReading *>(el_d);

    if (tl->name == Or) {
        int size = tl->tuple->contents.size();

        for (int i = 0; i < size; i++) {
            Data *rv = subst(tl->tuple->contents[i], el);
            if (rv != Zero) return rv;
        }

        return Zero;        
    }
    
    Substitution *sub = new Substitution();

    if (tl->name == Not) {
        tl = dynamic_cast<FunctorReading *>(tl->tuple->contents[0]);

        Data *es_d = subst(tl, el);

        // If we don't match anything, match everything.
        if (es_d == Zero) {
            delete sub;
            return empty;
        }

        Substitution *es = dynamic_cast<Substitution *>(es_d);
        
        // Otherwise, return the antimatter version of es.

        sub->antimatter = true;

        for (int i = 0; i < SIZE; i++) {
            sub->entry[i] = es->entry[i];
        }

        goto discrim;
        
    }
    
    if (tl->name != el->name) {
        delete sub;
        return Zero;
    }

    {
        int size = tl->tuple->contents.size();

        for(int i = 0; i < size; i++) {
            Data *tld = tl->tuple->contents[i];
            Data *eld = el->tuple->contents[i];

            if (tld == Wildcard) {
                continue;
            }
        
            if (tld->user == -1) {
                entry_labels[serial] = tld;
                tld->user = serial++;
            }

            if (sub->entry[tld->user] == NULL) {
                sub->entry[tld->user] = eld;
            } else {
                if (sub->entry[tld->user] != eld) {
                    delete sub;
                    return Zero;
                }
            }        
        }
    }

discrim:
    
    Substitution *rv = sub_discrim[sub];
    if (! rv) {
        sub_discrim[sub] = sub;
        rv = sub;
    } else {
        delete sub;
    }

    return rv;
    
}

Data *merge_core(Substitution *sub1, Substitution *sub2) {

    Substitution *rv = new Substitution();

    for (int i = 0; i < SIZE; i++) {

        Data *e1 = sub1->entry[i];
        Data *e2 = sub2->entry[i];
        Data *r = NULL;
        
        if (!e1) {
            r = e2;
        } else if (!e2) {
            r = e1;
        } else if (e1 == e2) {
            r = e1;
        } else {
            delete rv;
            return Zero;
        }

        rv->entry[i] = r;
    }

    Substitution *rv1 = sub_discrim[rv];
    if (rv1) {
        delete rv;
        return rv1;
    }

    sub_discrim[rv] = rv;
    return rv;    
}


Data *merge(Data *t1, Data *t2) {

#if 0
    printf("MERGE: ");
    t1->print();
    printf(" ");
    t2->print();
    printf("\n");
#endif
    
    if (t1 == Zero || t2 == Zero)
        return Zero;

    Substitution *sub1 = dynamic_cast<Substitution *>(t1);
    Substitution *sub2 = dynamic_cast<Substitution *>(t2);

    Data *rv = merge_core(sub1, sub2);

    if ((! sub1->antimatter) && (! sub2->antimatter)) {
        return rv;
    }

    if (sub1->antimatter && rv == Zero) {
        return sub2;
    }

    if (sub2->antimatter && rv == Zero) {
        return sub1;
    }

    return Zero;
}


Data *fwd_translate(Data *sub_d, Data *trans_d) {

    Substitution *sub = dynamic_cast<Substitution *>(sub_d);
    TupleReading *trans = dynamic_cast<TupleReading *>(trans_d);

    if (! sub) return Zero;
    
    Substitution *rv = new Substitution();

    for (int i = 0; i < SIZE; i++) {
        if (sub->entry[i]) {
            rv->entry[i] = Unreachable;
        } else {
            rv->entry[i] = NULL;
        }
    }
    
    for (int i = 0; i < trans->contents.size(); i += 2) {

        Data *out = trans->contents[i];
        Data *in = trans->contents[i + 1];


        for (int j = 0; i < SIZE; i++) {
            if (sub->entry[j] == out) {
                rv->entry[j] = in;
            }
        }
    }
    
    Substitution *rv1 = sub_discrim[rv];
    if (rv1) {
        delete rv;
        return rv1;
    }

    sub_discrim[rv] = rv;
    return rv;    
}


Data *rev_translate(Data *sub_d, Data *trans_d, Data *orig_d) {

    Substitution *sub = dynamic_cast<Substitution *>(sub_d);
    TupleReading *trans = dynamic_cast<TupleReading *>(trans_d);
    Substitution *orig = dynamic_cast<Substitution *>(orig_d);

    if (! sub || ! orig_d) return Zero;
    
    Substitution *rv = new Substitution();

    for (int i = 0; i < SIZE; i++) {

        if (sub->entry[i] == Unreachable) {
            rv->entry[i] = orig->entry[i];
        } else if (sub->entry[i]) {
            rv->entry[i] = Unreachable;
        } else {
            rv->entry[i] = NULL;
        }
    }
    
    for (int i = 0; i < trans->contents.size(); i += 2) {

        Data *out = trans->contents[i];
        Data *in = trans->contents[i + 1];

        for (int j = 0; i < SIZE; i++) {
            if (sub->entry[j] == in) {
                rv->entry[j] = out;
            }
        }
    }
    
    Substitution *rv1 = sub_discrim[rv];
    if (rv1) {
        delete rv;
        return rv1;
    }

    sub_discrim[rv] = rv;
    return rv;    
}


