#ifndef HASHING2_H
#define HASHING2_H

#ifndef PYAPTS_H
#error Hashing2 must be called from pyapts.
#endif

#include <functional>
#include <ext/hash_map>
using namespace std;
using namespace __gnu_cxx;


// This is the simplest version of NAMap. It's a map from nothing to a
// value. Since there's only one nothing, we have a single location,
// which we use to store the data.
template <class R, PNP> struct HashingMap0 {

    R *val;

    HashingMap0() {
        val = NULL;
    }
    
    R **lookup() {
        return &val;
    }
    
    R **lookup_create() {
        return &val;
    }
};


template <class A, class B>
struct T {

    A a;
    B b;

    ~T() {
    }
    
    T() {
    }
    
    T(const A a_in, const B b_in) {
        a = a_in;
        b = b_in;
    }

    T(const T &t) {
        a = t.a;
        b = t.b;
    }    
};

template <class A, class B>
T<A, B> t(A a, B b) {
    return T<A, B>(a, b);
}

// Hash functions.

static size_t hash2_hash(TypeBase *tbp) {
    return tbp->id;
}

template <class A, class B>
    size_t hash2_hash(const T<A, B> &t) {
    return 12347 * hash2_hash(t.a) + hash2_hash(t.b);
}

struct hash2_hasher {
    template <class A>
        size_t operator() (const A &a) const {

        return hash2_hash(a);
    }
};


// Equality functions.

static bool hash2_equal(const TypeBase *a, const TypeBase *b) {
    return a == b;
}

template <class A, class B>
    bool hash2_equal(const T<A, B> &left, const T<A, B> &right) {


    return hash2_equal(left.a, right.a) && hash2_equal(left.b, right.b);
}

struct hash2_equaler {
    template <class A, class B>
        size_t operator() (const A &a, const B &b) const {

        return hash2_equal(a, b);
    }
};

// HashingLookup Core.
    
template <class M, class K, class R>
    R *lookup_core(M &map, const K &key, R *empty) {
    
    typename M::iterator i = map.find(key);

    if (i == map.end()) {
        return empty;
    }

    return &((*i).second);
}

template <class P1, class R, PNP> struct HashingMap1 {
    hash_map<P1 *, R *, hash2_hasher, hash2_equaler> contents;

    R *empty;

    HashingMap1() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1) {
        return lookup_core(contents, p1, &empty);
    }

    R **lookup_create(P1 *p1) {
        return &(contents[p1]);
    }        
};
    

template <class P1, class P2, class R, PNP> struct HashingMap2 {
    hash_map<T<P1 *, P2 *>, R *, hash2_hasher, hash2_equaler> contents;

    R *empty;

    HashingMap2() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1, P2 *p2) {
        return lookup_core(contents, t(p1, p2), &empty);
    }

    R **lookup_create(P1 *p1, P2 *p2) {
        return &(contents[t(p1, p2)]);
    }        
};
    
template <class P1, class P2, class P3, class R, PNP> struct HashingMap3 {
    hash_map<T<P1 *, T<P2 *, P3 *> >, R *, hash2_hasher, hash2_equaler> contents;

    R *empty;

    HashingMap3() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1, P2 *p2, P3 *p3) {
        return lookup_core(contents, t(p1, t(p2, p3)), &empty);
    }

    R **lookup_create(P1 *p1, P2 *p2, P3 *p3) {
        return &(contents[t(p1, t(p2, p3))]);
    }        
};
    
template <class P1, class P2, class P3, class P4, class R, PNP> struct HashingMap4 {
    hash_map<T<T<P1 *, P2 *>, T<P3 *, P4 *> >, R *, hash2_hasher, hash2_equaler> contents;

    R *empty;

    HashingMap4() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1, P2 *p2, P3 *p3, P4 *p4) {
        return lookup_core(contents, t(t(p1, p2), t(p3, p4)), &empty);
    }

    R **lookup_create(P1 *p1, P2 *p2, P3 *p3, P4 *p4) {
        return &(contents[t(t(p1, p2), t(p3, p4))]);
    }        
};
    
template <class P1, class P2, class P3, class P4, class P5, class R, PNP> struct HashingMap5 {
    hash_map<T<T<P1 *, P2 *>, T<P3 *, T<P4 *, P5 *> > >, R *, hash2_hasher, hash2_equaler> contents;

    R *empty;

    HashingMap5() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1, P2 *p2, P3 *p3, P4 *p4, P5 *p5) {
        return lookup_core(contents, t(t(p1, p2), t(p3, t(p4, p5))), &empty);
    }

    R **lookup_create(P1 *p1, P2 *p2, P3 *p3, P4 *p4, P5 *p5) {
        return &(contents[t(t(p1, p2), t(p3, t(p4, p5)))]);
    }        
};
    



#endif
