#ifndef PYAPTS_H
#define PYAPTS_H

#include "data.h"
#include "reading.h"

#include <ext/hash_map>
#include <malloc.h>

#define PNP R** R::**pnp

using namespace __gnu_cxx;

struct TypeBase {
    int id;
};

// This is the simplest version of NAMap. It's a map from nothing to a
// value. Since there's only one nothing, we have a single location,
// which we use to store the data.
template <class R, PNP> struct NAMap0 {

    R *val;

    NAMap0() {
        val = NULL;
    }
    
    R **lookup() {
        return &val;
    }
    
    R **lookup_create() {
        return &val;
    }
};

//////////////////////////////////////////////////////////////////////////////

// Many hashes.
#ifdef HASHING
#include "hashing.h"
#endif

// 1 rts, rest hashed.
#include "new_hashing2.h"

// all hashed.
#ifdef HASHING4
#include "hashing4.h"
#endif


//////////////////////////////////////////////////////////////////////////////


template <class P, class R> struct NALookup {
    R **store;
    int size;

    NALookup() {
        size = 0;
        store = NULL;
    }
    
    R **lookup(P *p1, R **empty) {
        if (p1->id >= size) {
            return empty;
        }

        return &(store[p1->id]);
    }
    
    R **lookup_create(P *p1) {

        if (p1->id >= size) {
            int newsize = p1->id * 2 + 1;
            R **newstore = new R *[newsize];

            if (size) {
            
                for (int i = 0; i < size; i++) {
                    newstore[i] = store[i];
                }

            }

            for (int i = size; i < newsize; i++) {
                newstore[i] = NULL;
            }
                
            store = newstore;
            size = newsize;            

#if 0
            printf("--> %p\n", this);
        
            for (int i = 0; i < size; i++) {
                printf("%d %p\n", i, store[i]); 
            }

            printf("--\n");
#endif        


        }


        return &(store[p1->id]);
    }
    
};


template <class P, class R, PNP> struct NAFinalLookup {
    R **store;
    int size;

    NAFinalLookup() {
        size = 0;
        store = NULL;
    }
    
    R **lookup(P *p1, R **empty) {
        if (p1->id >= size) {
            return empty;
        }

        return &(store[p1->id]);
    }
    
    R **lookup_create(P *p1) {

        if (p1->id >= size) {
            int newsize = p1->id * 2 + 1;
            R **newstore = new R *[newsize];

            if (size) {
            
                for (int i = 0; i < size; i++) {
                    newstore[i] = store[i];
                }

            }

            if (size && *pnp) {
                for (int i = 0; i < size; i++) {
                    if (store[i]) {
                        // For debugging purposes.
                        // printf("PNP: %p %p\n", store[i]->**pnp, &store[i]);

                        store[i]->**pnp = &newstore[i];
                    }
                }
            }
            

            for (int i = size; i < newsize; i++) {
                newstore[i] = NULL;
            }
                
            store = newstore;
            size = newsize;            

#if 0
            printf("--> %p\n", this);
        
            for (int i = 0; i < size; i++) {
                printf("%d %p\n", i, store[i]); 
            }

            printf("--\n");
#endif        


        }


        return &(store[p1->id]);
    }
    
};

template <class P1, class R, PNP> struct NAMap1 {
    R *empty;
    NAFinalLookup<P1, R, pnp> l;

    void debug_dump() {
        printf("Debug Dump!!!!\n");
        for (int i = 0; i < l.size; i++) {
            printf("%d %p\n", i, l.store[i]);
        }
    }
    
    NAMap1() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1) {
        return l.lookup(p1, &empty);
    }

    R **lookup_create(P1 *p1) {
        return l.lookup_create(p1);
    }
};

template <class P1, class P2, class R, PNP> struct NAMap2 {

    R *empty;

    typedef NAFinalLookup<P2, R, pnp> ILType;
    NALookup<P1, ILType> l;
    
    NAMap2() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1, P2 *p2) {
        ILType **il1 = l.lookup(p1, NULL);

        if (il1 == NULL || *il1 == NULL) {
            return &empty;
        }

        return (*il1)->lookup(p2, &empty);
    }

    R **lookup_create(P1 *p1, P2 *p2) {
        ILType **il1 = l.lookup_create(p1);

        if (! *il1) {
            *il1 = new NAFinalLookup<P2, R, pnp>();
        }

        return (*il1)->lookup_create(p2);
    }
};

template <class P1, class P2, class P3, class R, PNP> struct NAMap3 {

    R *empty;

    typedef NAFinalLookup<P3, R, pnp> IL2Type;
    typedef NALookup<P2, IL2Type> IL1Type;
    NALookup<P1, IL1Type> l;
    
    NAMap3() {
        empty = NULL;
    }
    
    R **lookup(P1 *p1, P2 *p2, P3 *p3) {
        IL1Type **il1 = l.lookup(p1, NULL);

        if (il1 == NULL || *il1 == NULL) {
            return &empty;
        }

        IL2Type **il2 = (*il1)->lookup(p2, NULL);

        if (il2 == NULL || *il2 == NULL) {
            return &empty;
        }
        
        return (*il2)->lookup(p3, &empty);
    }

    R **lookup_create(P1 *p1, P2 *p2, P3 *p3) {

        IL1Type **il1 = l.lookup_create(p1);

        if (! *il1) {
            *il1 = new NALookup<P2, IL2Type>();
        }

        IL2Type **il2 = (*il1)->lookup_create(p2);

        if (! *il2) {
            *il2 = new NAFinalLookup<P3, R, pnp>();
        }
        
        return (*il2)->lookup_create(p3);
    }
};


// Profiling support.

void show_profile(const char *label);

#endif
