parser Parser:

set indent = "[ \t]", "#"

token ident		: '[a-zA-Z_]\w*';
token integer		: '[0-9]+';
token NL		: '\n';
token string            : '"[^"]*"';

separator whitespace	: '[ \r\t]+';
separator comment	: '#.*';



expr/e -> 
	expr10/e
	;

expr10/e -> 
	expr20/e 
	( "or" ! exp20/right {{ e = or_expr(left=e, right=right) }} )*
	;

expr20/e ->
	expr30/e 
	( "and" ! expr30/right {{ e = and_expr(left=e, right=right) }} )*
	;

expr30/e ->
	( "not" ! expr30/expr {{ e = not_expr(expr=expr) }}
	| expr40/e
	)
	;

expr40/e ->
	expr50/e
	( "in" ! expr50/s {{ e = in_expr(expr=e, set=s) }}
	| "not" "in" ! expr50/s {{ e = notin_expr(expr=e, set=s) }}
	)?
	;

expr50/e -> 
	expr100/e
	;

expr100/e -> 
	expr110/e
#	( ( '\+'/op | '-'/op ) expr110/right
#		{{ e = binary_expr(left=e, op=op, right=right) }} )*
	;

expr110/e ->
	expr190/e
	;

expr190/e ->
	expr200/e {{ exprs = [ e ] }} 
	  ( ( ',' expr200/e {{ exprs.append(e) }}  )+ 
	    {{ e = tuple_expr(exprs=exprs) }} )?
	;
    

expr200/e -> 
	( ident/name "\(" !
	  {{ e = call_expr(name=name, args=[]) }}

	  ( expr200/p {{ e.args.append(p) }}
            ( ',' expr200/p {{ e.args.append(p) }} )*
	  )? 
	"\)" 

	| expr210/e

	)
	;

expr210/e -> 
	expr300/e
	;

expr300/e ->
	( '\(' expr/e {{ e = paren_expr(expr=e) }} '\)'
	| integer/i {{ e = integer_expr(value=int(i)) }}
	| ident/name {{ e = variable_expr(name=name) }}
	| '%' {{ depth = 1 }}
	  ( '%' {{ depth = depth + 1 }} )*
          ident/name 
	  {{ e = subst_expr(name=name, depth=depth) }}

	)
	;

statements/block ->
	{{ block = [] }}
	( statement/s 
	{{ 
	   if s:
	       block.append(s)
	}}
	)+
	;

replacement/r ->
	"replace" ! ":" NL
		indent statement/pattern deindent
	"with" ":" NL
		block/replacement

	{{ r = replace_stmt(pattern=pattern, replacement=replacement) }}
	;

block/block ->
        (NL)*
	indent 
	statements/block
	deindent
	;

statement/s -> 
	( replacement/s
	| "invariant" ! expr/left "=" ! expr/right NL
		{{ s = invariant_stmt(left=left, right=right) }}
	| "@+"/depth ! NL
		{{ s = subst_stmt(depth=len(depth)) }}

	| expr/se "add" ! expr/expr NL
		{{ s = add_stmt(set=se, expr=expr) }}
	| expr/se "remove" ! expr/expr NL
		{{ s = remove_stmt(set=se, expr=expr) }}

        | "delete" expr/e NL
                {{ s = delete_stmt(expr=e) }}

	| "while" ! expr/expr ":" NL block/block 
		{{ s = while_stmt(expr=expr, block=block) }}
	| "if" ! expr/expr ":" NL block/block
		{{ s = if_stmt(expr=expr, block=block, else_block=[]) }}
	  ( "else" ":" NL block/else_block
		{{ s.else_block = else_block }} )?
	| "for" ! expr/expr ":" NL block/block
		{{ s = for_stmt(expr=expr, block=block) }}

	| expr/left "=" ! expr/right NL
		{{ s = assign_stmt(left=left, right=right) }}

	| "read" ! expr/e NL 
		{{ s = read_stmt(expr=e) }}

	| "print" ! expr/e NL 
		{{ s = print_stmt(expr=e) }}

        | "abort" ! ident/msg NL
                {{ s = abort_stmt(msg=msg) }}
        | "debug" ! ident/msg expr/e NL
                {{ s = debug_stmt(msg=msg, expr=e) }}
        | "smap" ! expr/e NL
                {{ s = smap_stmt(expr=e) }}
        | "profile" ! ident/msg NL
                {{ s = profile_stmt(msg=msg) }}
	| NL
		{{ s = None }}
	)
	;

START/s -> 
	statements/block
	{{ s = program(block=block) }}
	;
