class NoMatch(Exception):
    pass

class syntax(object):
    """Base class for syntax tree members.

This class provides a base for the members of the abstract
syntax tree.
"""

    def __init__(self, **dict):
        self.__dict__ = dict
        self.type = None

    def match(self, s, subst):
        
        if not self.__class__ == s.__class__:
            raise NoMatch

        for i in self.__class__.fields:
            a = getattr(self, i)
            b = getattr(s, i)
            
            if isinstance(a, str) or isinstance(a, int):
                if a != b:
                    raise NoMatch
                
            elif isinstance(a, list):
                for i, j in zip(a, b):
                    subst = i.match(j, subst)
                
            elif issubclass(a.__class__, syntax):
                subst = a.match(b, subst)

            else:
                if a is not b:
                    raise NoMatch

        return subst

    def duplicate(self, subst):

        c = self.__class__
        rv = c()
        
        for i in c.fields:
            a = getattr(self, i)

            if isinstance(a, str) or isinstance(a, int):
                setattr(rv, i, a)
                
            elif isinstance(a, list):

                b = []
                
                for j in a:
                    b.append(j.duplicate(subst))

                setattr(rv, i, b)

            else:
                setattr(rv, i, a.duplicate(subst))

        return rv

    def walk(self, pre_method, post_method, *args):

        if pre_method:
            getattr(self, pre_method)(*args)

        c = self.__class__

        for i in c.fields:
            a = getattr(self, i)

            if issubclass(a.__class__, syntax):
                a.walk(pre_method, post_method, *args)
            elif issubclass(a.__class__, list):
                for j in a:
                    j.walk(pre_method, post_method, *args)

        if post_method:
            getattr(self, post_method)(*args)


    def visit(self, method, *args):
        
        c = self.__class__

        for i in c.fields:
            a = getattr(self, i)

            if issubclass(a.__class__, syntax):
                getattr(a, method)(*args)
            elif issubclass(a.__class__, list):
                for j in a:
                    getattr(j, method)(*args)
        
            
    def find_type(self, tmap):
        self.type = None

    # This is used to compute the set of variables that are bound at
    # each point in the program, and with that, the strength of each
    # access to a set.
    def bindings(self, p):
        self.visit('bindings', p)
    
    def assign_merged_types(self, p):
        if self.type:
            self.type = self.type.resolve_types(p.tymap).merged()
            
