# This code is a mess, and needs to be rewritten. Basically, what this
# does is to start with a large number of types, nearly one per syntax
# tree node.
#
# We also have tymap, which is a map from variables to a list of
# types of that variable.
#
# We simplify this by merging types and resolving types, as appropriate.
# Types are merged if they're referred to by the same variable, by the same
# member of a tuple, by being the same parameter of a function, or by being
# the result of a function. Resolving types eliminates variable references.
#
# The result is a type DAG. Each type is represented exactly once. The
# type DAG no longer contains variable_types
#
# After this, we go through the syntax tree to be sure that every type
# annotation on that tree points to the single canonical type for that
# expression.

def working_on(v):
    if _working_on == v:
        return True
    else:
        return False

# If one side is None, we take the other side. Otherwise, we have to
# recursively merge.
def merge_types_core(a, b, tymap, special):

    if not b:
        return a

    if not a:
        return b

    a = a.merged()
    b = b.merged()

    if a is b:
        return a

#    print "merging", a, "with", b

    # Try the merge both ways, and take the one that works.
    am = a.merge(b, tymap, special)
    if am:
#        print "took a:", am
        return am

    bm = b.merge(a, tymap, special)
#    print "took b:", bm
    return bm

##############################################################################

class typing(object):

    def __init__(self, **dict):
        self.__dict__ = dict
        self.merged_with = None
    
    # This returns some sort of canonical version of this type. Specifically,
    # this eliminates one or more variable references.
    # def resolve(self, tymap):
    #    return self

    # The number of elements in this tuple. If we're not a tuple type,
    # only one element is needed. 
    def len(self):
        return 1


    # This returns a tuple of types.
    def type_tuple(self, tymap):
        return [ self ]

    def walk_types(self, m, *args):
        pass

    # This returns the type this type has been merged with. It does this
    # recursively, so that chains of merged types are resolved into a
    # single type.
    def merged(self):
        if self.merged_with:
#             print  "+ merged:", self, self.merged_with
            return self.merged_with.merged()

        return self

    # This is responsible for resolving self in the tymap. This alters
    # the current object so that it points to the canonical representation
    # for each variable.
    def resolve_types(self, tymap):
        assert 0

    # This gets all the types inside this type.
    def gather_types(self):

        rv = [ self ]

        for f in self.__class__.fields:
            v = getattr(self, f)

            if isinstance(v, list) or isinstance(v, tuple):
                for vv in v:
                    rv += vv.gather_types()

            if isinstance(v, typing):
                rv += v.gather_types()

        return rv
    
##############################################################################

## Hm... seems we may not need this.

# # This is a type that only exists to be merged with other types. It's
# # the default type 
# class unknown_type(typing):

#     def __str__(self):
#         return "unknown"

#     # If we have no information about the type of an expression,
#     # we can merge it with any other type.
#     def merge(self, o, tymap, special):
#         self.merged_with = o
#         return o


class function_type(typing):

    fields = [ 'args',
               'result',
               ]

    def __str__(self):
        return "(" + ", ".join([str(i) for i in self.args]) + ") -> " + str(self.result)

    def merge(self, o, tymap, special):
        
        if not isinstance(o, function_type):
            return None

        rv = function_type()

        rv.result = merge_types_core(self.result, o.result, tymap, special)

        rv.args = [ merge_types_core(a, b, tymap, special)
                    for a, b in zip(self.args, o.args) ]

        if rv.result == None or None in rv.args:
            return None

        self.merged_with = rv
        o.merged_with = rv

        return rv

    def walk_types(self, m, *args):        
        m(self, *args)
        self.result.walk_types(m, *args)
        for i in self.args:
            i.walk_types(m, *args)


    def resolve_types(self, tymap):
        self.result = self.result.merged().resolve_types(tymap)
        self.args = [ i.merged().resolve_types(tymap)
                      for i in self.args ]

        return self

    def struct_equiv(self, o):
        return self is o


##############################################################################

class int_type(typing):

    fields = []

    def __str__(self):
        return "int<" + str(id(self)) + ">"

    def walk_types(self, m, *args):
        m(self, *args)

    # We can merge with other int_types, only.
    def merge(self, o, tymap, special):

        if not isinstance(o, int_type):
            return None

        o.merged_with = self

        return self

    def resolve_types(self, tymap):
        return self

    def struct_equiv(self, o):
        return self is o

##############################################################################

class variable_type(typing):

    fields = [
        'name',
        ]

    def __str__(self):
        return self.name

    # If we have two variables, we merge them. Otherwise, we merge whatever
    # else we have into the variable.
    def merge(self, o, tymap, special):

#        print "In variable.merge!", str(self), str(o)

        # We only hande the case where o is a special, not the case
        # where self is a special.

        if self.name in special:
            return None


        if isinstance(o, variable_type):

#            print "Merging two variables."

            # Handle specials.
            if o.name in special:
                special[o.name].append(self.name)
                return self

            slist = tymap[self.name]
            olist = tymap[o.name]

            # If we're already merged, we have nothing to do. So, just
            # return self.
            if slist is olist:
                return self

#            print "Merging", self.name, "with", o.name, "."

            # Otherwise, we need to merge the variable lists.

            slist.extend(olist)

            # Now, ensure that all the types point the the merged
            # list. (With any luck, this will drop the references
            # to the other list from tymap.)

            for i in tymap:
                if tymap[i] is olist:
                    tymap[i] = slist
                    
            # Otherwise, we're done.

            return self

        # If the other side is a type, the variable can point to
        # the type.

#        print "In variable-other merge."

#        print "Added", o, "to", self.name
        tymap[self.name].append(o)

#        print "tymap of ", self.name, "is", tymap[self.name]


        # We don't want to do merged_with here, we do it later.

        return self

    # Variable type resolution step.
    def resolve_types(self, tymap):
        return tymap[self.name][0]

    def struct_equiv(self, o):
        assert False
        
##############################################################################

class tuple_type(typing):

    fields = [
        'members',
        ]

    def __str__(self):
        return "tuple(" + ", ".join([str(x) for x in self.members]) + ")"

    def len(self):
        return len(self.members)

    def type_tuple(self, tymap):
        return self.members
        
    # We can merge a tuple if we can merge all of the components
    # of the tuple.
    def merge(self, o, tymap, special):

        # The other side has to be a tuple.
        if not isinstance(o, tuple_type):
            return None

        sm = self.members
        om = o.members

        # The two tuples must be of the same length.

        if len(sm) != len(om):
            return None

        merged = [
            # The merge of the two members.
            merge_types_core(a, b, tymap, special)

            # For every pair of elements in the two lists.
            for a, b in zip(sm, om) ]

        # If we couldn't do the merge, there's at least one instance
        # of None in the list. So, we can use that to determine if
        # we succeded.

        if None in merged:
            return None

        # Otherwise, return the newly-merged tuple.
        rv = tuple_type(members = merged)

        self.merged_with = rv
        o.merged_with = rv

        return rv

    def walk_types(self, m, *args):
        m(self, *args)

        for t in self.members:
            t.walk_types(m, *args)

    def resolve_types(self, tymap):
        self.members = [ i.merged().resolve_types(tymap)
                         for i in self.members ]
        return self

    def struct_equiv(self, o):
        if not isinstance(o, tuple_type):
            return False

        if len(self.members) != len(o.members):
            return False

        for a, b in zip(self.members, o.members):
            if not a.struct_equiv(b):
                return False

        return True
                     

##############################################################################

class set_type(typing):

    fields = [
        'contains'
        ]

    def __str__(self):
        return "set(" + str(self.contains) + ")" 

    # We can merge the sets if we can merge the members of the set.
    def merge(self, o, tymap, special):

        # Check the types.
        if not isinstance(o, set_type):
            return None

        a = self.contains
        b = o.contains

        m = merge_types_core(a, b, tymap, special)

        if not m:
            return None

        rv = set_type(contains = m)
        self.merged_with = rv
        o.merged_with = rv

        return rv

    def walk_types(self, m, *args):
        m(self, *args)
        self.contains.walk_types(m, *args)

    def resolve_types(self, tymap):
        self.contains = self.contains.merged().resolve_types(tymap)
        return self

    def struct_equiv(self, o):
        if not isinstance(o, set_type):
            return False

        return self.contains.struct_equiv(o.contains)
    
##############################################################################

class typemap(dict):

    def put(self, key, val):
        if val:
            self.setdefault(key, []).append(val)

    # This trys to merge together types.
    def merge_types(self, special):

        vars = self.keys()

        modified = True

        while modified:
        
            modified = False

#             print 
#             print "===================="
#             print "Starting a new pass."
#             print "===================="

            for v in vars:

                global _working_on
                _working_on = v

#                print
#                print "Now Handling:", v
                
                types = self[v][:]

#                print "types:", ", ".join([str(i) for i in types])

                
                if len(types) < 2:
#                    print "Nothing to do for", v                    
                    continue

                modified = True
                rep = types[0].merged()

                for b in types[1:]:
                    b = b.merged()

                    if b is rep:
                        continue
                        
                    oldrep = rep
                    rep = merge_types_core(oldrep, b, self, special)

#                    print "FOOBAR", v, ":", oldrep, "merge", b, "=", rep

                    if not rep:
#                       print "Couldn't merge", oldrep, "with", b, "for var", v, "."
                        return None

                self[v][:] = [ rep ]
#                print v, "has new representation", rep


        # If we haven't determined a unique type for a variable, then
        # we just assume it takes an integer. Otherwise, resolve the
        # types reachable.
        for v in self:
            if len(self[v]) == 0:
                self[v].append(int_type())


        for v in self:
            self[v][:] = [ self[v][0].merged().resolve_types(self) ]
                    
        # Okay, at this point, we have uniqified tymap. For each variable v,
        # self[v][0] is the canonical type of that variable.


        # Now, handle structural merging. If two variables point to something
        # that is structurally the same, we should merge them into the same
        # thing. Two types are strucutrally the same if they have the same
        # leaf nodes, and the interior nodes are of the same types.

        # To make life a bit easier for us, we create a pair of data
        # structures. We create a list of leaders of merged variables,
        # and a map from variable names to the sets of variables they've
        # been merged with.

        leaders = []
        merged_vars = {}
        
        # First, handle normal variables. First find a leader for an
        # equivalence class...
        for v in self:

            if v in merged_vars:
                continue

            leaders.append(v)

            equivalent = [ ]

            # Then add in all the members of that class.
            for v1 in self:

                if self[v] is self[v1]:
                    equivalent.append(v1)
                    merged_vars[v1] = equivalent

        # Finally, add in specials.
        for v, list in special.iteritems():

            for v1 in list:
                if v not in merged_vars[v1]:
                    merged_vars[v1].append(v)

        # Remove the entry for None.
        if None in self:
            del self[None]

        return leaders, merged_vars

    def walk_types(self, m, *args):
        for v in self:
            t = self[v][0]
            t.walk_types(m, *args)

import os

if __name__ == "__main__":
    os.system("python main.py prpq.txt")
